# $ANTLR 3.5.2 cadbiom/models/biosignal/translators/sigexpr_lexer.g 2018-12-14 16:41:04

import sys
from antlr3 import *
from antlr3.compat import set, frozenset



# for convenience in actions
HIDDEN = BaseRecognizer.HIDDEN

# token types
EOF=-1
AND=4
CHG=5
COM=6
COMMENT=7
CONSTR=8
DEF=9
DEFAULT=10
DIGIT=11
DOL=12
EG=13
EVENT=14
EXC=15
EXP=16
F=17
IDENT=18
INC=19
LETTER=20
MINUS=21
MUL=22
NOT=23
NOTEG=24
OR=25
PD=26
PG=27
PLUS=28
SCOL=29
SEQ=30
SYNC=31
T=32
WHEN=33
WS=34


class sigexpr_lexer(Lexer):

    grammarFileName = "cadbiom/models/biosignal/translators/sigexpr_lexer.g"
    api_version = 1

    def __init__(self, input=None, state=None):
        if state is None:
            state = RecognizerSharedState()
        super(sigexpr_lexer, self).__init__(input, state)

        self.delegates = []

        self.dfa3 = self.DFA3(
            self, 3,
            eot = self.DFA3_eot,
            eof = self.DFA3_eof,
            min = self.DFA3_min,
            max = self.DFA3_max,
            accept = self.DFA3_accept,
            special = self.DFA3_special,
            transition = self.DFA3_transition
            )






    # $ANTLR start "WS"
    def mWS(self, ):
        try:
            _type = WS
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:53:13: ( ( ' ' | '\\t' | '\\n' ) )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:53:16: ( ' ' | '\\t' | '\\n' )
            pass 
            if (9 <= self.input.LA(1) <= 10) or self.input.LA(1) == 32:
                self.input.consume()
            else:
                mse = MismatchedSetException(None, self.input)
                self.recover(mse)
                raise mse



            #action start
            _channel = HIDDEN;
            #action end




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "WS"



    # $ANTLR start "COMMENT"
    def mCOMMENT(self, ):
        try:
            _type = COMMENT
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:55:13: ( '//' (~ '\\n' )* '\\n' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:55:15: '//' (~ '\\n' )* '\\n'
            pass 
            self.match("//")


            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:55:19: (~ '\\n' )*
            while True: #loop1
                alt1 = 2
                LA1_0 = self.input.LA(1)

                if ((0 <= LA1_0 <= 9) or (11 <= LA1_0 <= 65535)) :
                    alt1 = 1


                if alt1 == 1:
                    # cadbiom/models/biosignal/translators/sigexpr_lexer.g:
                    pass 
                    if (0 <= self.input.LA(1) <= 9) or (11 <= self.input.LA(1) <= 65535):
                        self.input.consume()
                    else:
                        mse = MismatchedSetException(None, self.input)
                        self.recover(mse)
                        raise mse




                else:
                    break #loop1


            self.match(10)

            #action start
            _channel = HIDDEN;
            #action end




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "COMMENT"



    # $ANTLR start "DEFAULT"
    def mDEFAULT(self, ):
        try:
            _type = DEFAULT
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:58:11: ( 'default' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:58:13: 'default'
            pass 
            self.match("default")




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "DEFAULT"



    # $ANTLR start "WHEN"
    def mWHEN(self, ):
        try:
            _type = WHEN
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:59:9: ( 'when' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:59:11: 'when'
            pass 
            self.match("when")




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "WHEN"



    # $ANTLR start "EVENT"
    def mEVENT(self, ):
        try:
            _type = EVENT
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:60:9: ( 'event' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:60:11: 'event'
            pass 
            self.match("event")




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "EVENT"



    # $ANTLR start "SYNC"
    def mSYNC(self, ):
        try:
            _type = SYNC
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:61:9: ( 'synchro' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:61:11: 'synchro'
            pass 
            self.match("synchro")




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "SYNC"



    # $ANTLR start "EXC"
    def mEXC(self, ):
        try:
            _type = EXC
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:62:9: ( 'exclus' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:62:11: 'exclus'
            pass 
            self.match("exclus")




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "EXC"



    # $ANTLR start "INC"
    def mINC(self, ):
        try:
            _type = INC
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:63:9: ( 'included' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:63:13: 'included'
            pass 
            self.match("included")




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "INC"



    # $ANTLR start "SEQ"
    def mSEQ(self, ):
        try:
            _type = SEQ
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:64:9: ( 'sequence' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:64:13: 'sequence'
            pass 
            self.match("sequence")




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "SEQ"



    # $ANTLR start "AND"
    def mAND(self, ):
        try:
            _type = AND
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:65:9: ( 'and' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:65:11: 'and'
            pass 
            self.match("and")




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "AND"



    # $ANTLR start "OR"
    def mOR(self, ):
        try:
            _type = OR
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:66:9: ( 'or' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:66:11: 'or'
            pass 
            self.match("or")




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "OR"



    # $ANTLR start "NOT"
    def mNOT(self, ):
        try:
            _type = NOT
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:67:9: ( 'not' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:67:11: 'not'
            pass 
            self.match("not")




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "NOT"



    # $ANTLR start "T"
    def mT(self, ):
        try:
            _type = T
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:68:9: ( 'true' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:68:11: 'true'
            pass 
            self.match("true")




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "T"



    # $ANTLR start "F"
    def mF(self, ):
        try:
            _type = F
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:69:9: ( 'false' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:69:11: 'false'
            pass 
            self.match("false")




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "F"



    # $ANTLR start "CONSTR"
    def mCONSTR(self, ):
        try:
            _type = CONSTR
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:70:9: ( 'constraint' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:70:13: 'constraint'
            pass 
            self.match("constraint")




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "CONSTR"



    # $ANTLR start "DEF"
    def mDEF(self, ):
        try:
            _type = DEF
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:72:9: ( ':=' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:72:13: ':='
            pass 
            self.match(":=")




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "DEF"



    # $ANTLR start "EG"
    def mEG(self, ):
        try:
            _type = EG
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:73:9: ( '=' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:73:13: '='
            pass 
            self.match(61)



            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "EG"



    # $ANTLR start "NOTEG"
    def mNOTEG(self, ):
        try:
            _type = NOTEG
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:74:9: ( '!=' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:74:11: '!='
            pass 
            self.match("!=")




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "NOTEG"



    # $ANTLR start "PG"
    def mPG(self, ):
        try:
            _type = PG
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:75:9: ( '(' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:75:11: '('
            pass 
            self.match(40)



            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "PG"



    # $ANTLR start "PD"
    def mPD(self, ):
        try:
            _type = PD
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:76:9: ( ')' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:76:11: ')'
            pass 
            self.match(41)



            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "PD"



    # $ANTLR start "DOL"
    def mDOL(self, ):
        try:
            _type = DOL
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:77:9: ( '$' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:77:11: '$'
            pass 
            self.match(36)



            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "DOL"



    # $ANTLR start "COM"
    def mCOM(self, ):
        try:
            _type = COM
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:78:9: ( ',' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:78:11: ','
            pass 
            self.match(44)



            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "COM"



    # $ANTLR start "UP"
    def mUP(self, ):
        try:
            _type = UP
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:79:9: ( '>' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:79:13: '>'
            pass 
            self.match(62)



            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "UP"



    # $ANTLR start "DOWN"
    def mDOWN(self, ):
        try:
            _type = DOWN
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:80:9: ( '<' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:80:13: '<'
            pass 
            self.match(60)



            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "DOWN"



    # $ANTLR start "CHG"
    def mCHG(self, ):
        try:
            _type = CHG
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:81:9: ( '!' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:81:13: '!'
            pass 
            self.match(33)



            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "CHG"



    # $ANTLR start "SCOL"
    def mSCOL(self, ):
        try:
            _type = SCOL
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:82:9: ( ';' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:82:13: ';'
            pass 
            self.match(59)



            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "SCOL"



    # $ANTLR start "PLUS"
    def mPLUS(self, ):
        try:
            _type = PLUS
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:83:9: ( '+' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:83:13: '+'
            pass 
            self.match(43)



            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "PLUS"



    # $ANTLR start "MINUS"
    def mMINUS(self, ):
        try:
            _type = MINUS
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:84:9: ( '-' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:84:13: '-'
            pass 
            self.match(45)



            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "MINUS"



    # $ANTLR start "MUL"
    def mMUL(self, ):
        try:
            _type = MUL
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:85:9: ( '*' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:85:13: '*'
            pass 
            self.match(42)



            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "MUL"



    # $ANTLR start "EXP"
    def mEXP(self, ):
        try:
            _type = EXP
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:86:9: ( '^' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:86:13: '^'
            pass 
            self.match(94)



            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "EXP"



    # $ANTLR start "LETTER"
    def mLETTER(self, ):
        try:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:89:19: ( 'a' .. 'z' | 'A' .. 'Z' | '_' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:
            pass 
            if (65 <= self.input.LA(1) <= 90) or self.input.LA(1) == 95 or (97 <= self.input.LA(1) <= 122):
                self.input.consume()
            else:
                mse = MismatchedSetException(None, self.input)
                self.recover(mse)
                raise mse






        finally:
            pass

    # $ANTLR end "LETTER"



    # $ANTLR start "DIGIT"
    def mDIGIT(self, ):
        try:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:91:19: ( '0' .. '9' )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:
            pass 
            if (48 <= self.input.LA(1) <= 57):
                self.input.consume()
            else:
                mse = MismatchedSetException(None, self.input)
                self.recover(mse)
                raise mse






        finally:
            pass

    # $ANTLR end "DIGIT"



    # $ANTLR start "IDENT"
    def mIDENT(self, ):
        try:
            _type = IDENT
            _channel = DEFAULT_CHANNEL

            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:95:9: ( ( LETTER | DIGIT )+ )
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:95:11: ( LETTER | DIGIT )+
            pass 
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:95:11: ( LETTER | DIGIT )+
            cnt2 = 0
            while True: #loop2
                alt2 = 2
                LA2_0 = self.input.LA(1)

                if ((48 <= LA2_0 <= 57) or (65 <= LA2_0 <= 90) or LA2_0 == 95 or (97 <= LA2_0 <= 122)) :
                    alt2 = 1


                if alt2 == 1:
                    # cadbiom/models/biosignal/translators/sigexpr_lexer.g:
                    pass 
                    if (48 <= self.input.LA(1) <= 57) or (65 <= self.input.LA(1) <= 90) or self.input.LA(1) == 95 or (97 <= self.input.LA(1) <= 122):
                        self.input.consume()
                    else:
                        mse = MismatchedSetException(None, self.input)
                        self.recover(mse)
                        raise mse




                else:
                    if cnt2 >= 1:
                        break #loop2

                    eee = EarlyExitException(2, self.input)
                    raise eee

                cnt2 += 1




            self._state.type = _type
            self._state.channel = _channel
        finally:
            pass

    # $ANTLR end "IDENT"



    def mTokens(self):
        # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:8: ( WS | COMMENT | DEFAULT | WHEN | EVENT | SYNC | EXC | INC | SEQ | AND | OR | NOT | T | F | CONSTR | DEF | EG | NOTEG | PG | PD | DOL | COM | UP | DOWN | CHG | SCOL | PLUS | MINUS | MUL | EXP | IDENT )
        alt3 = 31
        alt3 = self.dfa3.predict(self.input)
        if alt3 == 1:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:10: WS
            pass 
            self.mWS()



        elif alt3 == 2:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:13: COMMENT
            pass 
            self.mCOMMENT()



        elif alt3 == 3:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:21: DEFAULT
            pass 
            self.mDEFAULT()



        elif alt3 == 4:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:29: WHEN
            pass 
            self.mWHEN()



        elif alt3 == 5:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:34: EVENT
            pass 
            self.mEVENT()



        elif alt3 == 6:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:40: SYNC
            pass 
            self.mSYNC()



        elif alt3 == 7:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:45: EXC
            pass 
            self.mEXC()



        elif alt3 == 8:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:49: INC
            pass 
            self.mINC()



        elif alt3 == 9:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:53: SEQ
            pass 
            self.mSEQ()



        elif alt3 == 10:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:57: AND
            pass 
            self.mAND()



        elif alt3 == 11:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:61: OR
            pass 
            self.mOR()



        elif alt3 == 12:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:64: NOT
            pass 
            self.mNOT()



        elif alt3 == 13:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:68: T
            pass 
            self.mT()



        elif alt3 == 14:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:70: F
            pass 
            self.mF()



        elif alt3 == 15:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:72: CONSTR
            pass 
            self.mCONSTR()



        elif alt3 == 16:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:79: DEF
            pass 
            self.mDEF()



        elif alt3 == 17:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:83: EG
            pass 
            self.mEG()



        elif alt3 == 18:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:86: NOTEG
            pass 
            self.mNOTEG()



        elif alt3 == 19:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:92: PG
            pass 
            self.mPG()



        elif alt3 == 20:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:95: PD
            pass 
            self.mPD()



        elif alt3 == 21:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:98: DOL
            pass 
            self.mDOL()



        elif alt3 == 22:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:102: COM
            pass 
            self.mCOM()



        elif alt3 == 23:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:106: UP
            pass 
            self.mUP()



        elif alt3 == 24:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:109: DOWN
            pass 
            self.mDOWN()



        elif alt3 == 25:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:114: CHG
            pass 
            self.mCHG()



        elif alt3 == 26:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:118: SCOL
            pass 
            self.mSCOL()



        elif alt3 == 27:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:123: PLUS
            pass 
            self.mPLUS()



        elif alt3 == 28:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:128: MINUS
            pass 
            self.mMINUS()



        elif alt3 == 29:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:134: MUL
            pass 
            self.mMUL()



        elif alt3 == 30:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:138: EXP
            pass 
            self.mEXP()



        elif alt3 == 31:
            # cadbiom/models/biosignal/translators/sigexpr_lexer.g:1:142: IDENT
            pass 
            self.mIDENT()








    # lookup tables for DFA #3

    DFA3_eot = DFA.unpack(
        u"\3\uffff\13\34\2\uffff\1\101\14\uffff\1\34\2\uffff\1\34\2\uffff"
        u"\2\34\2\uffff\2\34\2\uffff\1\34\2\uffff\1\34\2\uffff\1\132\2\uffff"
        u"\1\34\2\uffff\1\34\2\uffff\1\34\2\uffff\1\34\52\uffff"
        )

    DFA3_eof = DFA.unpack(
        u"\150\uffff"
        )

    DFA3_min = DFA.unpack(
        u"\1\11\2\uffff\1\145\1\150\1\166\1\145\2\156\1\162\1\157\1\162\1"
        u"\141\1\157\2\uffff\1\75\14\uffff\1\146\2\uffff\1\145\2\uffff\1"
        u"\145\1\143\2\uffff\1\156\1\161\2\uffff\1\143\2\uffff\1\144\2\uffff"
        u"\1\60\2\uffff\1\164\2\uffff\1\165\2\uffff\1\154\2\uffff\1\156\52"
        u"\uffff"
        )

    DFA3_max = DFA.unpack(
        u"\1\172\2\uffff\1\145\1\150\1\170\1\171\2\156\1\162\1\157\1\162"
        u"\1\141\1\157\2\uffff\1\75\14\uffff\1\146\2\uffff\1\145\2\uffff"
        u"\1\145\1\143\2\uffff\1\156\1\161\2\uffff\1\143\2\uffff\1\144\2"
        u"\uffff\1\172\2\uffff\1\164\2\uffff\1\165\2\uffff\1\154\2\uffff"
        u"\1\156\52\uffff"
        )

    DFA3_accept = DFA.unpack(
        u"\1\uffff\1\1\1\2\13\uffff\1\20\1\21\1\uffff\1\23\1\24\1\25\1\26"
        u"\1\27\1\30\1\32\1\33\1\34\1\35\1\36\1\37\43\uffff\1\22\1\31\1\3"
        u"\2\uffff\1\4\2\uffff\1\5\2\uffff\1\7\2\uffff\1\6\2\uffff\1\11\2"
        u"\uffff\1\10\2\uffff\1\12\2\uffff\1\13\1\uffff\1\14\2\uffff\1\15"
        u"\2\uffff\1\16\2\uffff\1\17\2\uffff"
        )

    DFA3_special = DFA.unpack(
        u"\150\uffff"
        )


    DFA3_transition = [
        DFA.unpack(u"\2\1\25\uffff\1\1\1\20\2\uffff\1\23\3\uffff\1\21\1\22"
        u"\1\32\1\30\1\24\1\31\1\uffff\1\2\12\34\1\16\1\27\1\26\1\17\1\25"
        u"\2\uffff\32\34\3\uffff\1\33\1\34\1\uffff\1\10\1\34\1\15\1\3\1\5"
        u"\1\14\2\34\1\7\4\34\1\12\1\11\3\34\1\6\1\13\2\34\1\4\3\34"),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u"\1\35"),
        DFA.unpack(u"\1\40"),
        DFA.unpack(u"\1\43\1\uffff\1\44"),
        DFA.unpack(u"\1\50\23\uffff\1\47"),
        DFA.unpack(u"\1\53"),
        DFA.unpack(u"\1\56"),
        DFA.unpack(u"\1\61"),
        DFA.unpack(u"\1\64"),
        DFA.unpack(u"\1\67"),
        DFA.unpack(u"\1\72"),
        DFA.unpack(u"\1\75"),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u"\1\100"),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u"\1\102"),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u"\1\105"),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u"\1\110"),
        DFA.unpack(u"\1\113"),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u"\1\116"),
        DFA.unpack(u"\1\121"),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u"\1\124"),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u"\1\127"),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u"\12\34\7\uffff\32\34\4\uffff\1\34\1\uffff\32\34"),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u"\1\134"),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u"\1\137"),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u"\1\142"),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u"\1\145"),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u""),
        DFA.unpack(u"")
    ]

    # class definition for DFA #3

    class DFA3(DFA):
        pass


 



def main(argv, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr):
    from antlr3.main import LexerMain
    main = LexerMain(sigexpr_lexer)

    main.stdin = stdin
    main.stdout = stdout
    main.stderr = stderr
    main.execute(argv)



if __name__ == '__main__':
    main(sys.argv)



# -*- coding: utf-8 -*-
## Filename    : TestMCLTranslators.py
## Author(s)   : Michel Le Borgne
## Created     : 9 mars 2012
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s):
##
"""
Unitary tests for the translators
"""
from __future__ import print_function
import unittest
import sys

from cadbiom.models.clause_constraints.CLDynSys import CLDynSys
from cadbiom.models.clause_constraints.mcl.MCLTranslators import  MCLSigExprVisitor, \
            gen_transition_clock, gen_transition_list_clock, \
            gen_simple_evolution, GT2Clauses
from cadbiom.models.biosignal.sig_expr import SigIdentExpr, \
            SigSyncBinExpr, SigWhenExpr, SigDefaultExpr
from cadbiom.models.guard_transitions.chart_model import ChartModel
from cadbiom.models.guard_transitions.analyser.ana_visitors import TableVisitor
from cadbiom.models.clause_constraints.mcl.MCLAnalyser import MCLAnalyser


# simple reporter
class ErrorRep(object):
    """
    Simple error reporter
    """
    def __init__(self):
        self.mess = ""
        self.error = False
        pass

    def display(self, mess):
        """
        set error and print
        """
        self.error = True
        self.mess += '\n>> '+mess
        print('\n DISPLAY CALL>> '+mess)

TRACE_FILE = sys.stdout
#TRACE_FILE = open("/tmp/testMCLTranslator.txt",'w')

class TestTransitionClauses(unittest.TestCase):
    """
    Test of transitions into clauses
    """

    def test_no_cond(self): #OK
        """
        n1 --> n2; []
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_simple_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        trans = root.add_transition(nn1, nn2)

        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys(tvisit, reporter)
        TRACE_FILE.write('\ntestNoCond')
        TRACE_FILE.write( '\n\nn1 --> n2; []\n')
        htr = gen_transition_clock(trans, cl_ds, None, reporter)
        TRACE_FILE.write( 'Transition clock:'+ htr.__str__()+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        mess = 'free clocks registered:' + cl_ds.free_clocks.__str__() + '\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'reporter: '+ reporter.mess+'\n')

    def test_cond(self): #OK
        """
        n1 --> n2; [not n3]
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_simple_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        nn3 = root.add_simple_node('n3', 0, 0)
        trans = root.add_transition(nn1, nn2)
        trans.set_condition("not n3")

        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, None)
        TRACE_FILE.write('\ntestCond')
        TRACE_FILE.write( '\n\nn1 --> n2; [not n3]'+'\n')
        htr = gen_transition_clock(trans, cl_ds, None, reporter)
        TRACE_FILE.write( 'Transition clock:' + htr.__str__()+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__() + '\n')
        mess =  'free clocks registered:' + cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'reporter: '+ reporter.mess+'\n')

    def test_no_cond_event(self): #OK
        """
        n1 --> n2; h[]
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_simple_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        trans = root.add_transition(nn1, nn2)
        trans.set_event('h')

        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys(tvisit.tab_symb, reporter)
        TRACE_FILE.write('\ntestNoCondEvent')
        TRACE_FILE.write( '\n\nn1 --> n2; h[]'+'\n')
        htr = gen_transition_clock(trans, cl_ds, None, reporter)
        TRACE_FILE.write( 'Transition clock:'+ htr.__str__()+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__() + '\n')
        mess = 'free clocks registered:' +cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'reporter: '+ reporter.mess+'\n')

        TRACE_FILE.write( '---------- opti ----------------\n')
        sed = dict()
        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        cl_ds = CLDynSys(tvisit.tab_symb, reporter)
        htr = gen_transition_clock(trans, cl_ds, sed, reporter)
        TRACE_FILE.write( 'Transition clock:'+ htr.__str__()+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        mess = 'free clocks registered:' + cl_ds.free_clocks.__str__() + '\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'reporter: '+ reporter.mess+'\n')

    def test_cond_event(self): #OK
        """
        n1 --> n2; h[not n3]
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_simple_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        nn3 = root.add_simple_node('n3', 0, 0)
        trans = root.add_transition(nn1, nn2)
        trans.set_condition("not n3")
        trans.set_event('h')

        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, None)
        TRACE_FILE.write('\ntestCondEvent')
        TRACE_FILE.write( '\n\nn1 --> n2; h[not n3]'+'\n')
        htr = gen_transition_clock(trans, cl_ds, None, reporter)
        TRACE_FILE.write( 'Transition clock:'+ htr.__str__()+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        mess = 'free clocks registered:'+ cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'reporter: '+ reporter.mess+'\n')

        TRACE_FILE.write( '---------- opti ----------------\n')
        sed = dict()
        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, None)
        TRACE_FILE.write('\ntestCondEvent')
        TRACE_FILE.write( '\n\nn1 --> n2; h[not n3]'+'\n')
        htr = gen_transition_clock(trans, cl_ds, sed, reporter)
        TRACE_FILE.write( 'Transition clock:'+ htr.__str__()+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        mess = 'free clocks registered:'+ cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'reporter: '+ reporter.mess+'\n')





    def test_perm_no_cond_event(self): #OK
        """
        n1/p --> n2; h[];
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_perm_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        trans = root.add_transition(nn1, nn2)
        trans.set_event('h')

        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, None)
        TRACE_FILE.write('\ntestPermNoCondEvent')
        TRACE_FILE.write( '\n\nn1/p --> n2; h[]'+'\n' )
        htr = gen_transition_clock(trans, cl_ds, None, reporter)
        TRACE_FILE.write( 'Transition clock:'+htr.__str__()+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        mess =  'free clocks registered:'+ cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'reporter: '+ reporter.mess+'\n')

    def test_perm_cond_event(self): #OK
        """
        n4;
        n1/p --> n2; h[n4];
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_perm_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        nn4 = root.add_simple_node('n4', 0, 0)
        trans = root.add_transition(nn1, nn2)
        trans.set_condition("n4")
        trans.set_event('h')

        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, None)
        TRACE_FILE.write('\ntestPermCondEvent')
        TRACE_FILE.write( '\n\nn1/p --> n2; h[n4]' +'\n')
        htr = gen_transition_clock(trans, cl_ds, None, reporter)
        TRACE_FILE.write( 'Transition clock:'+ htr.__str__()+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write(cla.__str__()+'\n')
        mess = 'free clocks registered:'+ cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'reporter: '+ reporter.mess+'\n')

    def test_input_cond_event(self): #OK
        """
        n4/i;
        n1/i --> n2; h[n4];
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_input_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        nn4 = root.add_input_node('n4', 0, 0)
        trans = root.add_transition(nn1, nn2)
        trans.set_condition("n4")
        trans.set_event('h')

        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, None)
        TRACE_FILE.write('\ntestInputCondEvent')
        TRACE_FILE.write( '\n\nn1/i --> n2; h[n4]'+'\n')
        htr = gen_transition_clock(trans, cl_ds, None, reporter)
        TRACE_FILE.write( 'Transition clock:' +htr.__str__()+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        mess = 'free clocks registered:'+ cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'reporter: '+ reporter.mess+'\n')

    # complex events
    def test_no_cond_event_when1(self): #OK
        """
        n1 --> n2; h when n3[]
        n3;
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_simple_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        nn3 = root.add_simple_node('n3', 0, 0)
        trans = root.add_transition(nn1, nn2)
        trans.set_event('h when n3')

        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys(tvisit.tab_symb, reporter)
        TRACE_FILE.write('\ntestNoCondEventWhen1')
        TRACE_FILE.write( '\n\nn1 --> n2; h when n3[]; n3;'+'\n')
        htr = gen_transition_clock(trans, cl_ds, None, reporter)
        TRACE_FILE.write( 'Transition clock:'+ htr.__str__()+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        mess = 'free clocks registered:' +cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'place_clocks'+ cl_ds.place_clocks.__str__()+'\n')
        TRACE_FILE.write( 'reporter: '+ reporter.mess+'\n')

    def test_no_cond_event_when2(self): #OK
        """
        n1 --> n2; h when n3[]
        n3 --> n1 ; h2 when h[]
        Error h is not a state
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_simple_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        nn3 = root.add_simple_node('n3', 0, 0)
        tr1 = root.add_transition(nn1, nn2)
        tr1.set_event('h when n3')
        tr2 = root.add_transition(nn3, nn1)
        tr2.set_event('h2 when h')

        reporter = ErrorRep()
        tvisit = TableVisitor(reporter) # no error display
        model.accept(tvisit)

        cl_ds = CLDynSys(tvisit.tab_symb, reporter)
        TRACE_FILE.write('\ntestNoCondEventWhen1')
        mess =  '\n\nn1 --> n2; h when n3[]; n3 --> n1 ; h2 when h[];'+'\n'
        TRACE_FILE.write(mess)
        htr = gen_transition_clock(tr1, cl_ds, None, reporter)
        TRACE_FILE.write( 'Transition clock:'+ htr.__str__()+'\n')
        htr = gen_transition_clock(tr2, cl_ds, None, reporter)
        TRACE_FILE.write( 'Transition clock:'+ htr.__str__()+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        mess =  'free clocks registered:' +cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'place_clocks'+ cl_ds.place_clocks.__str__()+'\n')
        TRACE_FILE.write( 'reporter: '+ reporter.mess+'\n')

    def test_no_cond_event_when3(self): #OK
        """
        n1 --> n2; h when n3[]
        n3 --> n1 ; n2 when n1[]
        Error n2 is not a clock
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_simple_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        nn3 = root.add_simple_node('n3', 0, 0)
        tr1 = root.add_transition(nn1, nn2)
        tr1.set_event('h when n3')
        tr2 = root.add_transition(nn3, nn1)
        tr2.set_event('n2 when n1')

        reporter = ErrorRep()
        tvisit = TableVisitor(reporter) # no error display
        model.accept(tvisit)

        cl_ds = CLDynSys(tvisit.tab_symb, reporter)
        TRACE_FILE.write('\ntestNoCondEventWhen1')
        TRACE_FILE.write( '\n\nn1 --> n2; n2 when n1[]; n3;'+'\n')
        htr = gen_transition_clock(tr1, cl_ds, None, reporter)
        TRACE_FILE.write( 'Transition clock:'+ htr.__str__()+'\n')
        htr = gen_transition_clock(tr2, cl_ds, None, reporter)
        TRACE_FILE.write( 'Transition clock:'+ htr.__str__()+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        mess = 'free clocks registered:' +cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'place_clocks'+ cl_ds.place_clocks.__str__()+'\n')
        TRACE_FILE.write( 'reporter: '+ reporter.mess+'\n')


    def test_no_cond_event_default(self): #OK
        """
        n1 --> n2; h1 when c1 default h2[c3]
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_simple_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        root.add_simple_node('c1', 0, 0)
        root.add_simple_node('c3', 0, 0)
        tr1 = root.add_transition(nn1, nn2)
        tr1.set_event('h when c1 default h2')
        tr1.set_condition('c3')

        reporter = ErrorRep()
        tvisit = TableVisitor(reporter) # no error display
        model.accept(tvisit)

        cl_ds = CLDynSys(tvisit.tab_symb, reporter)
        TRACE_FILE.write('\ntestNoCondEventWhen1')
        TRACE_FILE.write( '\n\nn1 --> n2; n2 when n1[]; n3;'+'\n')
        htr = gen_transition_clock(tr1, cl_ds, None, reporter)
        TRACE_FILE.write( 'Transition clock:'+ htr.__str__()+'\n')

        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        mess =  'free clocks registered:' +cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'place_clocks'+ cl_ds.place_clocks.__str__()+'\n')
        TRACE_FILE.write( 'reporter: '+ reporter.mess+'\n')


class TestTransitionList(unittest.TestCase):
    """
    As it says
    """

#    def setUp(self):
#        self.output = open('TestTransition.txt')
#        #self.output = sys.stdout


    def test_no_cond_on_out2events(self):
        """
        n1 --> n2; h1[not n4]
        n1 --> n3; h2[]
        n4 --> n1; h3[]
        gen_transition_list_clock tested
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_simple_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        nn3 = root.add_simple_node('n3', 0, 0)
        nn4 = root.add_simple_node('n4', 0, 0)
        trans = root.add_transition(nn1, nn2)
        trans.set_condition("not n4")
        trans.set_event('h1')
        trans = root.add_transition(nn1, nn3)
        trans.set_event('h2')
        trans = root.add_transition(nn4, nn1)
        trans.set_event('h3')

        tvisit = TableVisitor(None) # no error display/ might crash
        model.accept(tvisit)
        TRACE_FILE.write('\ntestNoCondOnOut2Events')
        mess = '\n\nn1 --> n2; h1[not n4] \nn1 --> n3; h2[]\nn4 --> n1; h3[]\n'
        TRACE_FILE.write(mess)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, None)
        trl = nn1.outgoing_trans
        sed = dict()
        htr = gen_transition_list_clock(trl, cl_ds, sed, reporter)
        TRACE_FILE.write( 'Transition clock:' +htr.__str__()+'\n')
        TRACE_FILE.write( 'NB Clauses:' +str(len(cl_ds.list_clauses))+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        mess = 'free clocks registered:'+cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)

    def test_very_simple(self): #OK to optimize
        """
        n1 --> n2; []
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_simple_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        trans = root.add_transition(nn1, nn2)


        model.clean_code()
        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, None)
        TRACE_FILE.write('\ntestVerySimple')
        TRACE_FILE.write( '\n\nn1 --> n2; [] \n')
        gen_simple_evolution(nn1, cl_ds, None, reporter)
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        mess =  'free clocks registered:'+cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        model.clean_code()

    def test_simple(self):
        """
        n1 --> n2; h1[not n4]
        n1 --> n3; h2[]
        n4 --> n1; h3[]
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_simple_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        nn3 = root.add_simple_node('n3', 0, 0)
        nn4 = root.add_simple_node('n4', 0, 0)
        trans = root.add_transition(nn1, nn2)
        trans.set_condition("not n4")
        trans.set_event('h1')
        trans = root.add_transition(nn1, nn3)
        trans.set_event('h2')
        trans = root.add_transition(nn4, nn1)
        trans.set_event('h3')

        model.clean_code()
        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, None)
        TRACE_FILE.write('\ntestSimple')
        mess =  '\n\nn1 --> n2; h1[not n4] \nn1 --> n3; h2[]\nn4 --> n1; h3[]\n'
        TRACE_FILE.write(mess)
        sed = dict()
        gen_simple_evolution(nn1, cl_ds, sed, reporter)
        TRACE_FILE.write( 'NB Clauses:' + str(len(cl_ds.list_clauses)) + '\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__() + '\n')
        mess = 'free clocks registered:'+ cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        model.clean_code()

    def test_simple_in_no_out(self):
        """
        n2 --> n1; h1[not n4]
        n4 --> n1; h3[]
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_simple_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        nn3 = root.add_simple_node('n3', 0, 0)
        nn4 = root.add_simple_node('n4', 0, 0)
        trans = root.add_transition(nn2, nn1)
        trans.set_condition("not n4")
        trans.set_event('h1')
        trans = root.add_transition(nn4, nn1)
        trans.set_event('h3')

        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, reporter)
        TRACE_FILE.write('\ntestSimpleInNoOut')
        TRACE_FILE.write( '\n\nn2 --> n1; h1[not n4] \nn4 --> n1; h3[]\n')
        gen_simple_evolution(nn1, cl_ds, None, reporter)
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        mess =  'free clocks registered:' + cl_ds.free_clocks.__str__() + '\n'
        TRACE_FILE.write(mess)

    def test_simple_out_no_in(self):
        """
        n1 --> n2; h1[not n4]
        n1 --> n3; h2[]
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_simple_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        nn3 = root.add_simple_node('n3', 0, 0)
        nn4 = root.add_simple_node('n4', 0, 0)
        trans = root.add_transition(nn1, nn2)
        trans.set_condition("not n4")
        trans.set_event('h1')
        trans = root.add_transition(nn1, nn3)
        trans.set_event('h2')

        model.clean_code()
        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, None)
        TRACE_FILE.write('\ntestSimpleOutNoIn')
        TRACE_FILE.write( '\n\nn1 --> n2; h1[not n4] \nn1 --> n3; '+'\n')
        gen_simple_evolution(nn1, cl_ds, None, reporter)
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        mess =  'free clocks registered:'+ cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)

    def test_simple_no_trans(self): #OK
        """
        n1;
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_simple_node('n1', 0, 0)

        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, None)
        TRACE_FILE.write('\ntestSimpleNoTrans')
        TRACE_FILE.write( '\n\nn1; \n')
        gen_simple_evolution(nn1, cl_ds, None, reporter)
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__() + '\n')
        mess = 'free clocks registered:' + cl_ds.free_clocks.__str__() + '\n'
        TRACE_FILE.write(mess)


class TestFull(unittest.TestCase):
    """
    test full models
    """

    def test_model1(self):
        """
        n1 --> n2; h1[not n4]
        n1 --> n3; h2[]
        n4 --> n1; h3[]
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_simple_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        nn3 = root.add_simple_node('n3', 0, 0)
        nn4 = root.add_simple_node('n4', 0, 0)
        trans = root.add_transition(nn1, nn2)
        trans.set_condition("not n4")
        trans.set_event('h1')
        trans = root.add_transition(nn1, nn3)
        trans.set_event('h2')
        trans = root.add_transition(nn4, nn1)
        trans.set_event('h3')

        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, reporter)
        TRACE_FILE.write('\ntestModel1')
        mess = '\n\nn1 --> n2; h1[not n4] \nn1 --> n3; h2[]\nn4 --> n1; h3[]\n'
        TRACE_FILE.write(mess)
        gt2clauses = GT2Clauses(cl_ds, False, reporter)
        model.accept(gt2clauses)
        TRACE_FILE.write( 'NB Clauses:' +str(len(cl_ds.list_clauses))+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        TRACE_FILE.write( 'variables'+ cl_ds.base_var_set.__str__()+'\n')
        mess = 'free clocks registered:'+ cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'place_clocks'+ cl_ds.place_clocks.__str__()+'\n')
        TRACE_FILE.write( 'inputs'+ cl_ds.inputs.__str__()+'\n')

    def test_model2(self): #OK
        """
        n1 --> n2; h1 when n2 default h2[not n4]
        n1 --> n3; h2 when n4[]
        n4 --> n1; h3[]
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_simple_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        nn3 = root.add_simple_node('n3', 0, 0)
        nn4 = root.add_simple_node('n4', 0, 0)
        trans = root.add_transition(nn1, nn2)
        trans.set_condition("not n4")
        trans.set_event('h1 when n2 default h2')
        trans = root.add_transition(nn1, nn3)
        trans.set_event('h2 when n4')
        trans = root.add_transition(nn4, nn1)
        trans.set_event('h3')

        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, reporter)
        TRACE_FILE.write('\ntestModel2')
        mess =  '\n\nn1 --> n2; h1 when n2 default h2[not n4]'
        mess = mess + '\nn1 --> n3; h2 when n4[]\nn4 --> n1; h3[]\n'
        TRACE_FILE.write(mess)
        gt2clauses = GT2Clauses(cl_ds, reporter, False)
        model.accept(gt2clauses)
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        TRACE_FILE.write( 'variables'+ cl_ds.base_var_set.__str__()+'\n')
        mess = 'free clocks registered:'+ cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'place_clocks'+ cl_ds.place_clocks.__str__()+'\n')
        TRACE_FILE.write( 'inputs'+ cl_ds.inputs.__str__()+'\n')


    def test_model3(self): #OK
        """
        n4;
        n1/p --> n2; h[n4];
        """
        model = ChartModel("Test")
        root = model.get_root()
        nn1 = root.add_perm_node('n1', 0, 0)
        nn2 = root.add_simple_node('n2', 0, 0)
        nn4 = root.add_simple_node('n4', 0, 0)
        trans = root.add_transition(nn1, nn2)
        trans.set_condition("n4")
        trans.set_event('h')

        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, None)
        TRACE_FILE.write('\ntestInputCondEvent')
        TRACE_FILE.write( '\n\nn4; \nn1/p --> n2; h[n4]'+'\n')

        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, reporter)
        TRACE_FILE.write('\ntestModel2')
        TRACE_FILE.write( '\n\n\n')
        gt2clauses = GT2Clauses(cl_ds, reporter, False)
        model.accept(gt2clauses)
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        TRACE_FILE.write( 'variables'+ cl_ds.base_var_set.__str__()+'\n')
        mess =  'free clocks registered:'+ cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'place_clocks'+ cl_ds.place_clocks.__str__()+'\n')
        TRACE_FILE.write( 'inputs'+ cl_ds.inputs.__str__()+'\n')


    def test_constraints(self):
        """
        As it says
        """
        model = ChartModel("Test")
        root = model.get_root()
        aaa = root.add_simple_node('A', 0, 0)
        bbb = root.add_simple_node('B', 0, 0)
        ccc = root.add_simple_node('C', 0, 0)
        ddd = root.add_simple_node('D', 0, 0)
        trans = root.add_transition(aaa, bbb)
        trans.set_event('h1 default h2')
        trans = root.add_transition(ccc, ddd)
        trans.set_event('h3')
        cont = 'synchro(h1, h3);\nexclus(h1, h2);\nincluded(h3, h2);'
        model.constraints = cont
        tvisit = TableVisitor(None) # no error display
        model.accept(tvisit)
        reporter = ErrorRep()
        cl_ds = CLDynSys( tvisit.tab_symb, reporter)
        gt2clauses = GT2Clauses(cl_ds, reporter, False)
        model.accept(gt2clauses)
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')
        TRACE_FILE.write( 'variables'+ cl_ds.base_var_set.__str__()+'\n')
        mess = 'free clocks registered:'+ cl_ds.free_clocks.__str__()+'\n'
        TRACE_FILE.write(mess)
        TRACE_FILE.write( 'place_clocks'+ cl_ds.place_clocks.__str__()+'\n')
        TRACE_FILE.write( 'inputs'+ cl_ds.inputs.__str__()+'\n')

    @unittest.skip("Test files not provided")
    def test_tgf_no_clock(self):
        """
        optimisation gain comparison
        """
        rep = ErrorRep()
        mcla = MCLAnalyser(rep)
        mcla.opti = False
        mcla.build_from_chart_file("../ucl/examples/test_tgfb_ref_300511.bcx")
        mess = '\n- NB Clauses:' + str(len(mcla.dynamical_system.list_clauses))
        TRACE_FILE.write(mess)
        mcla = MCLAnalyser(rep)
        mcla.build_from_chart_file("../ucl/examples/test_tgfb_ref_300511.bcx")
        mess = '\n- NB Clauses:' + str(len(mcla.dynamical_system.list_clauses))
        TRACE_FILE.write(mess)

class TestMCLSigExprVisitor (unittest.TestCase):
    """
    Test of signal expression
    """
    def test_ident1(self):
        """
        sigexp = xx
        """
        sexpr = SigIdentExpr('xx')
        cl_ds = CLDynSys(None, None)
        sexpv = MCLSigExprVisitor(cl_ds, 'Y', None)
        var = sexpr.accept(sexpv)
        TRACE_FILE.write( '\n\n sigexpr: xx'+'\n')
        TRACE_FILE.write( 'RETURN var (xx)'+ var.__str__()+'\n')
        TRACE_FILE.write( 'Clauses'+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')

    def test_bin1(self):
        """
        X or Y
        """
        sexpr1 = SigIdentExpr('x')
        sexpr2 = SigIdentExpr('y')
        sexpr = SigSyncBinExpr('or', sexpr1, sexpr2)
        cl_ds = CLDynSys(None, None)
        sexpv = MCLSigExprVisitor(cl_ds, 'Z', dict())
        var = sexpr.accept(sexpv)
        TRACE_FILE.write( '\n\n sigexpr: X or Y'+'\n')
        TRACE_FILE.write( 'RETURN var (Z)'+ var.__str__()+'\n')
        TRACE_FILE.write( 'Clauses'+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')

    def test_bin2(self):
        """
        X and Y
        """
        sexpr1 = SigIdentExpr('x')
        sexpr2 = SigIdentExpr('y')
        sexpr = SigSyncBinExpr('and', sexpr1, sexpr2)
        cl_ds = CLDynSys(None, None)
        sexpv = MCLSigExprVisitor(cl_ds, 'Z', dict())
        sexpv.output_lit = None # for output var generation
        var = sexpr.accept(sexpv)
        TRACE_FILE.writelines( '\n\n sigexpr: X and Y'+'\n')
        TRACE_FILE.write( 'RETURN var (aux var)'+ var.__str__()+'\n')
        TRACE_FILE.write( 'Clauses'+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')

    def test_bin3(self):
        """
        h1 default h2
        """
        sexpr1 = SigIdentExpr('h1')
        sexpr2 = SigIdentExpr('h2')
        sexpr = SigDefaultExpr(sexpr1, sexpr2)
        symb_tab = dict()
        symb_tab['h1'] = ('clock', -1)
        symb_tab['h2'] = ('clock', -1)
        cl_ds = CLDynSys(symb_tab, None)
        sexpv = MCLSigExprVisitor(cl_ds, None)
        #sexpv.output_lit = None # for output var generation
        var = sexpr.accept(sexpv)
        TRACE_FILE.writelines( '\n\n sigexpr: h1 default h2'+'\n')
        TRACE_FILE.write( 'RETURN var (aux var)'+ var.__str__()+'\n')
        TRACE_FILE.write( 'Clauses'+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')

    def test_bin4(self):
        """
        h1 when (a or b)
        """
        sexpr1 = SigIdentExpr('h1')
        sexpr2 = SigIdentExpr('h2')
        sexpr3 = SigIdentExpr('a')
        sexpr4 = SigIdentExpr('b')
        sexpr = SigSyncBinExpr('or', sexpr3, sexpr4)
        sexpr = SigWhenExpr(sexpr1, sexpr)
        symb_tab = dict()
        symb_tab['h1'] = ('clock', -1)
        symb_tab['h2'] = ('clock', -1)
        cl_ds = CLDynSys(symb_tab, None)
        sexpv = MCLSigExprVisitor(cl_ds, None)
        #sexpv.output_lit = None # for output var generation
        var = sexpr.accept(sexpv)
        TRACE_FILE.writelines( '\n\n sigexpr: h1 when (a or b)'+'\n')
        TRACE_FILE.write( 'RETURN var (aux var)'+ var.__str__()+'\n')
        TRACE_FILE.write( 'Clauses'+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')

    def test_cse1(self):
        """
        common subexpression elimination
        h1 when (a or b) default h2 when (b or a)
        """
        sexpr1 = SigIdentExpr('h1')
        sexpr2 = SigIdentExpr('h2')
        sexpr3 = SigIdentExpr('a')
        sexpr4 = SigIdentExpr('b')
        a_or_b1 = SigSyncBinExpr('and', sexpr3, sexpr4)
        a_or_b2 = SigSyncBinExpr('and', sexpr4, sexpr3)
        hh1 = SigWhenExpr(sexpr1, a_or_b1)
        hh2 = SigWhenExpr(sexpr2, a_or_b2)
        sexpr = SigDefaultExpr(hh1, hh2)
        TRACE_FILE.write('Formula: ' + sexpr.__str__())
        symb_tab = dict()
        symb_tab['h1'] = ('clock', -1)
        symb_tab['h2'] = ('clock', -1)
        cl_ds = CLDynSys(symb_tab, None)
        sub_exp = dict()
        sexpv = MCLSigExprVisitor(cl_ds, None, sub_exp)
        #sexpv.output_lit = None # for output var generation
        var = sexpr.accept(sexpv)
        TRACE_FILE.write( '\n\n- RETURN var (aux var)'+ var.__str__()+'\n')
        TRACE_FILE.write('\n- NB Clauses:' + str(len(cl_ds.list_clauses)))
        TRACE_FILE.write( '\n- Clauses'+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')

    def test_cse2(self):
        """
        common subexpression elimination
        h1 when (a or b) default h2 when (b or a)
        """
        hh1 = SigIdentExpr('h1')
        hh2 = SigIdentExpr('h2')
        h12 = SigDefaultExpr(hh1, hh2)
        h21 = SigDefaultExpr(hh2, hh1)
        sexpr = SigDefaultExpr(h12, h21)
        TRACE_FILE.write('Formula: ' + sexpr.__str__())
        symb_tab = dict()
        symb_tab['h1'] = ('clock', -1)
        symb_tab['h2'] = ('clock', -1)
        cl_ds = CLDynSys(symb_tab, None)
        sub_exp = dict()
        sexpv = MCLSigExprVisitor(cl_ds, None, sub_exp)
        #sexpv.output_lit = None # for output var generation
        var = sexpr.accept(sexpv)
        TRACE_FILE.write( '\n\n- RETURN var (aux var)'+ var.__str__()+'\n')
        TRACE_FILE.write('\n- NB Clauses:' + str(len(cl_ds.list_clauses)))
        TRACE_FILE.write( '\n- Clauses'+'\n')
        for cla in cl_ds.list_clauses:
            TRACE_FILE.write( cla.__str__()+'\n')



if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()

#    suiteFew = unittest.TestSuite()
    #suiteFew.addTest(TestMCLSigExprVisitor("testCSE2"))
    #suiteFew.addTest(Testgt2Sig("testOneTransCondEvent"))
    #suiteFew.addTest(TestTransitionClauses('testCondEvent'))
    #suiteFew.addTest(TestFull('testModel3'))
    #suiteFew.addTest(TestTransitionList("testSimple"))
    #unittest.TextTestRunner(verbosity=2).run(suiteFew)



# -*- coding: utf-8 -*-
## Filename    : MCLQuery.py
## Author(s)   : Michel Le Borgne
## Created     : 10 sept. 2012
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux - IRISA/IRSET
##
"""
Query internal representation
"""
from cadbiom.models.clause_constraints.mcl.MCLSolutions import MCLException
from cadbiom import commons as cm
from logging import DEBUG

LOGGER = cm.logger()


class MCLSimpleQuery(object):
    """Class packaging the elements of a query

    Object containing 2 main types of attributes describing properties:

        - Attributes in text format: These are logical formulas that are humanly
        readable.
            Ex: start_prop, inv_prop, final_prop, variant_prop
        - Attributes in DIMACS format: These are logical formulas encoded in
        the form of clauses containing numerical values.
            Ex: dim_start, dim_inv, dim_final, dim_variant_prop

    Textual properties of query are compiled into numeric clauses in the unfolder
    with init_forward_unfolding(), just at the beginning of squery_is_satisfied()
    and squery_solve().

    Attributes:

        :param start_prop: start property; logical formulas
        :param inv_prop: invariant property; logical formulas
        :param final_prop: final property; logical formulas
        :param variant_prop: list of logical formulas from ic_sequence.
            It's the trajectory of events of a solution.
            :Example:
                ['', 'h2 and h2', '', 'h2']

        :param dim_start: start property in DIMACS form - optional
        :param dim_inv: invariant property in DIMACS form - optional
        :param dim_final: final property in DIMACS form - optional
        :param dim_variant_prop: list of lists of dimacs clauses
        :param steps_before_check: Number of shifts before testing the
            final property - optional

        :type start_prop: <str>
        :type inv_prop: <str>
        :type final_prop: <str>
        :type variant_prop: <list <str>>

        :type dim_start: <list <DClause>>, default []
        :type dim_inv: <list <DClause>>, default []
        :type dim_final: <list <DClause>>, default []
        :type dim_variant_prop: <list <list <DClause>>
        :type steps_before_check: <int>, default value 0

    NB: DClause: A clause coded as a list of DIMACS coded literals: <list <int>>

    """

    def __init__(self, start_prop, inv_prop, final_prop):
        """
        @param start_prop: init property - None is allowed
        @param inv_prop: invariant property - None allowed
        @param final_prop: final property - None is allowed
        """
        if LOGGER.getEffectiveLevel() == DEBUG:
            LOGGER.debug(
                "MCLSimpleQuery params:: start prop: "
                + str(start_prop)
                + "; inv prop: "
                + str(inv_prop)
                + "; final prop: "
                + str(final_prop)
            )
        # Strings forms of constraints
        self.start_prop = start_prop  # logical formula or None
        self.inv_prop = inv_prop      # logical formula or None
        self.final_prop = final_prop  # logical formula or None
        self.variant_prop = []        # list<logical formula>

        # DIMACS forms of the preceding attributes
        self.dim_start = []         # list<DClause>
        self.dim_inv = []           # list<DClause>
        self.dim_final = []         # list<DClause>
        self.dim_variant_prop = []  # list<list<DClause>>

        self.steps_before_check = 0

    @classmethod
    def from_frontier_sol(cls, f_sol):
        """Build a query from a frontier solution

        Start property is based on activated frontiers.

        All frontiers that are not in activated_frontier
        **ARE NOT** forced to be inactivated with a negative value.

        Variant property enforces same timing on activated events,
        and the others are free.

        @param f_sol: FrontierSolution (human readable solution)
        @param unfolder: current unfolder
        """
        # start condition enforce activation of solution places
        start_prop = None
        if f_sol.activated_frontier:
            start_prop = " and ".join(f_sol.activated_frontier)

        # no invariant property
        inv_prop = None

        # no final property
        final_prop = None

        # variant property enforce same timing
        # Each event has a list of steps which it belongs
        # Ex: for steps: ['%', '% h2 h2', '%', '% h2']
        # variant property: ['', 'h2 and h2', '', 'h2']
        print("Reading ic_sequence from FrontierSolution...")

        # - Iterate on steps
        # - Get events names in each step
        # (remove the leading "%" and split the string on spaces)
        # - join events names with a logical operator " and "
        var_prop = [
            " and ".join(raw_step[1:].split()) for raw_step in f_sol.ic_sequence
        ]
        # PS: var_prop and any attribute can be [], but keep this for uniformity
        if not var_prop:
            var_prop = None

        n_query = MCLSimpleQuery(start_prop, inv_prop, final_prop)
        n_query.variant_prop = var_prop
        return n_query

    @classmethod
    def from_frontier_sol_new_timing(cls, f_sol, unfolder):
        """Build a query from a frontier solution

        Start property is based on activated frontiers.

        Search all frontiers that are not in activated_frontier
        and force their inactivation with a negative value.

        **Variant property enforces new timing**

        @param f_sol: FrontierSolution (human readable solution)
        @param unfolder: unfolder
        """
        # activation part of start condition
        start_prop = None
        if f_sol.activated_frontier:
            start_prop = " and ".join(f_sol.activated_frontier)

        # no invariant property
        inv_prop = None

        # no final property
        final_prop = None

        # variant property enforce new timing on activated events
        # Each event has a list of steps which it belongs
        # Ex: for steps: ['%', '% h2 h2', '%', '% h2']
        # variant property: ['', 'not (h2) or not (h2)', '', 'not (h2)']
        print("Reading ic_sequence from FrontierSolution...")

        # - Iterate on steps
        # - Get events names in each step
        # (remove the leading "%" and split the string on spaces)
        # - negation of each event with "not ( event )"
        # - join events names with a logical operator " or "
        var_prop = [
            " or ".join("not (" + icp + ")" for icp in raw_step[1:].split())
            for raw_step in f_sol.ic_sequence
        ]
        # PS: var_prop and any attribute can be [], but keep this for uniformity
        if not var_prop:
            var_prop = None

        # Inactivation of other frontier places at start (DIMACS format)
        # Search all frontiers that are not in activated_frontier
        # and force their inactivation with a negative value.
        ## dim_start doit-il etre ordonné ?
        # 1 - Get negative values of activated frontiers in FrontierSolution object
        activated_frontier_neg_values = {
            -unfolder.var_dimacs_code(name) for name in f_sol.activated_frontier
        }

        # 2 - Get values of inactivated frontiers in the FrontSolution object
        dim_start_values = \
            [[val] for val in unfolder.frontiers_negative_values - activated_frontier_neg_values]


        n_query = MCLSimpleQuery(start_prop, inv_prop, final_prop)
        n_query.variant_prop = var_prop
        n_query.dim_start = dim_start_values
        return n_query

    @classmethod
    def from_frontier_sol_same_timing(cls, f_sol, unfolder):
        """Build a query from a frontier solution

        Start property is based on activated frontiers.

        Search all frontiers that are not in activated_frontier
        and force their inactivation with a negative value.

        **Variant property enforces same timing on activated events,
        and the others are free.**

        @param f_sol: FrontierSolution (human readable solution)
        @param unfolder: current unfolder
        """
        # activation part of start condition
        start_prop = None
        if f_sol.activated_frontier:
            start_prop = " and ".join(f_sol.activated_frontier)

        # no invariant property
        inv_prop = None

        # no final property
        final_prop = None

        # variant property enforce same timing on activated events
        # Each event has a list of steps which it belongs
        # Ex: for steps: ['%', '% h2 h2', '%', '% h2']
        # variant property: ['', 'h2 and h2', '', 'h2']
        print("Reading ic_sequence from FrontierSolution...")

        # - Iterate on steps
        # - Get events names in each step
        # (remove the leading "%" and split the string on spaces)
        # - join events names with a logical operator " and "
        var_prop = [
            " and ".join(raw_step[1:].split()) for raw_step in f_sol.ic_sequence
        ]
        # PS: var_prop and any attribute can be [], but keep this for uniformity
        if not var_prop:
            var_prop = None

        # inactivation of other frontier places at start (DIMACS format)
        # Search all frontiers that are not in activated_frontier
        # and force their inactivation with a negative value.
        # Get negative values of activated frontiers in FrontierSolution object
        activated_frontier_neg_values = {
            -unfolder.var_dimacs_code(name) for name in f_sol.activated_frontier
        }

        # Get list of values of inactivated frontiers in the FrontSolution object
        dim_start_values = \
            [[val] for val in unfolder.frontiers_negative_values - activated_frontier_neg_values]


        n_query = MCLSimpleQuery(start_prop, inv_prop, final_prop)
        # list of logical formulas
        n_query.variant_prop = var_prop
        # start property in DIMACS format
        n_query.dim_start = dim_start_values
        return n_query

    def merge(self, query):
        """Return a new query which is a merge of two queries into one.

        - start, invariant and final properties are merged.
        - If both queries have variant properties, they must be on the same horizon
        (number of steps).
        - steps before reach is set to zero.
        - dim properties are also merged: dim_start, dim_inv, dim_final

        @param query: a MCLSimpleQuery
        @raise MCLException: if the queries have variant properties on
        different horizons.
        """
        # merge of start properties
        if not self.start_prop:
            start_prop = query.start_prop
        elif not query.start_prop:
            start_prop = self.start_prop
        else:
            start_prop = "(" + self.start_prop + ") and (" + query.start_prop + ")"
        # merge of invariant properties
        if not self.inv_prop:
            i_prop = query.inv_prop
        elif not query.inv_prop:
            i_prop = self.inv_prop
        else:
            i_prop = "(" + self.inv_prop + ") and (" + query.inv_prop + ")"
        # merge of final properties
        if not self.final_prop:
            final_prop = query.final_prop
        elif not query.final_prop:
            final_prop = self.final_prop
        else:
            final_prop = "(" + self.final_prop + ") and (" + query.final_prop + ")"

        n_query = MCLSimpleQuery(start_prop, i_prop, final_prop)

        # merge of variant prop
        var_prop = None
        if not self.variant_prop:
            var_prop = query.variant_prop
        elif not query.variant_prop:
            var_prop = self.variant_prop
        else:
            ll1 = len(self.variant_prop)
            ll2 = len(query.variant_prop)
            if ll1 != ll2:
                raise MCLException(
                    "Tempting to merge two queries with variant prop of different lengths"
                    "current: {}; other: {}".format(ll1, ll2)
                )
            else:
                var_prop = []
                # 2 empty, 1 empty and not the other, none empty
                for current, other in zip(self.variant_prop, query.variant_prop):
                    if current and other:
                        var_prop.append(current + " and " + other)
                    elif current and not other:
                        var_prop.append(current)
                    elif not current and other:
                        var_prop.append(other)
                    else:
                        var_prop.append("")

        n_query.variant_prop = var_prop

        # merge of dim properties
        n_query.dim_start += query.dim_start
        n_query.dim_inv += query.dim_inv
        n_query.dim_final += query.dim_final
        return n_query

    def __str__(self):
        """Print logical formulas in query"""
        str_out = "Start_property: "
        if self.start_prop:
            str_out = str_out + self.start_prop
        str_out = str_out + "\nInv property: "
        if self.inv_prop:
            str_out = str_out + self.inv_prop
        str_out = str_out + "\nFinal property: "
        if self.final_prop:
            str_out = str_out + self.final_prop
        str_out = str_out + "\nVariant property: "
        if self.variant_prop:
#            str_out = str_out + self.variant_prop[0]
#            for prop in self.variant_prop[1:]:
#                str_out = str_out + ", " + prop
            str_out = str_out + str(self.variant_prop)
        return str_out



## Filename    : chart_xml_pid.py
## Author(s)   : Geoffroy Andrieux
## Created     : 02/2012
## Revision    :
## Source      :
##
## Copyright 2010 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Geoffroy Andrieux.
##     IRISA/IRSET
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Michel Le Borgne
##
"""
Guarded transition interpretation of PID data
"""

from cadbiom.models.guard_transitions.chart_model import ChartModel, ChartModelException
from lxml import etree

class MEvent(object):
    """
    Object merging similar events in PID
    """
    clock_count = 0
    def __init__(self, inputs, outputs, act_list, inhib_list, pmid, ai_inter):
        """
        Create a merged event. Similar reactions (same inputs and outputs) exist in PID we merge them.
        @param inputs: list of strings - not void
        @param outputs: list of strings - not void
        @param act_list: list of strings
        @param inhib_list: list of strings
        @param pmid: reference
        @param ai_inter: interpretation of activators and inhibitors (0: one activator no inhib., 1: all activ. inhib if all inhib )
        """
        self.inputs = inputs
        self.inputs.sort()
        self.outputs = outputs
        self.outputs.sort()
        # a unique key corresponds to a combination of inputs - outputs
        self.key = self.inputs[0]
        for input in self.inputs[1:]:
            self.key = self.key + '!' + input
        self.key = self.key + '#'
        for output in self.outputs:
            self.key = self.key + '!' + output

        # compute condition as a logical formula (or or and)
        if ai_inter == 0:
            act_logic = logical_or(act_list)
            inhib_logic = logical_or(inhib_list)
        elif ai_inter == 1:
            act_logic = logical_and(act_list)
            inhib_logic = logical_and(inhib_list)
        else:
            raise ChartModelException("Unknown activator and inhibitor option")

        if inhib_logic is not None :
            inhib_logic = 'not('+inhib_logic+')'
        condition_list = []
        if act_logic is not None :
            condition_list.append(act_logic)
        if inhib_logic is not None :
            condition_list.append(inhib_logic)
        condition = logical_and(condition_list)
        self.condition = condition  # might be None
        self.pmid = pmid

        # clock is set from outside (set_clock)
        self.clock = '_$_' #indeterminate clock


    def set_clock(self):
        """
        Create a clock ident and store it in the object
        """
        self.clock = '_h_%i' % MEvent.clock_count
        MEvent.clock_count += 1

    def merge(self, mev):
        """
        merge two events:
        we take the "or" of the two conditions
        references are concatenated without redondancy check
        """
        if self.condition and mev.condition:
            cond1 = "(" + self.condition + ")"
            cond2 = "(" + mev.condition + ")"
            self.condition = cond1 + 'or' + cond2
        elif mev.condition:
            self.condition = mev.condition
        self.pmid = self.pmid + mev.pmid

    def __str__(self):
        str_out = '\nInputs: '
        for input in self.inputs:
            str_out = str_out + input + ", "
        str_out = str_out+"\nOutput: "
        for output in self.outputs:
            str_out = str_out + output + ", "
        str_out = str_out+"\nCondition: "
        if self.condition:
            str_out = str_out + self.condition
        str_out = str_out+"\nClock: "
        str_out = str_out + self.clock
        str_out = str_out +'\n'
        return str_out




class PidXmlParser:
    """
    Object for parsing a PID xml file
    """
    def __init__(self, xml_file, reporter, has_clock = True,
                 ai_inter=0, model=None, e_loc = []):
        """
        Do everything
        """
        self.reporter = reporter
        self.has_clock = has_clock
        self.ai_interpretation = ai_inter
        # statistics
        self.node_count = 0
        self.transition_count = 0
        self.transcription_count = 0
        self.extern_loc = e_loc # for statistics on "extern" molecules
        self.extern_count = 0

        #day=time.strftime('%d_%m_%y',time.localtime())
        if not model:
            xf_split = xml_file.split('.')
            self.model = ChartModel(xf_split[0])
        else:
            self.model = model
        self.top = self.model.get_root()
        self.dict_node = dict()
        self.dict_transition = dict()

        self.coord_x = 0.01
        self.coord_y = 0.01
        self.compt = 1

        self.dict_id = dict()               # id -> name for molecules
        self.dict_name = dict()             # name -> id for molecules
        self.page = etree.parse(xml_file)
        self.int_cpt = 0
        self.mol_cpt = 0
        self.location_dict = dict()
        self.make_dict()                    # build dict_id and dict_name
        dict_mev = self.make_mev_dict()     # build a temporary dictionary
        print "\n\n\nNB INTERACTIONS: "+str(self.int_cpt)
        print "\n\n\nNB MOLECULES: "+str(self.mol_cpt)
        cpt = 0
        for key in self.location_dict.keys() :
            cpt += self.location_dict[key]
            print key,'\t',self.location_dict[key]
        print cpt
#        for k in dict_mev.keys():
#            print dict_mev[k]
        for kmev in dict_mev:
            self.make_transition(dict_mev[kmev])
        # for statistics
        self.clock_count = MEvent.clock_count


    def get_model(self):
        """
        model accessor
        """
        return self.model

#    def set_extern_localisations(self, l_loc):
#        """
#        Define external localisations in PID
#        """
#        self.extern_loc = l_loc

    def make_dict(self):
        """
        build the two dictionaries dict_id and dict_name - Cadbiom basic names are created.
        Cadbiom names result from a mangling of basic name ('PF' one if any) and various informations
        such that localisation, activity, ptm ...see throught_components
        """
        mol_list = self.page.find('Model/MoleculeList')
        i = 0
        for mol in mol_list:
            mid = mol.get('id')                   # id nb in PID
            mtype = mol.get('molecule_type')
            mol_name_list = mol.findall('Name')     # list of synonyms in PID
            if mtype:
                for name in mol_name_list:
                    if name.get('name_type') == "PF":  # preferred name
                        mol_name = name.get('value')
                        i += 1
                        break
                    else :
                        # last synonym if no preferred name type
                        mol_name = name.get('value')

                if self.dict_id.has_key(mid) and self.dict_id[mid]!=mol_name:
                    mess = 'id error : '+str(mid)
                    mess = mess + ' has two different names : '
                    mess = mess + str(mol_name) + ' and '+str(self.dict_id[mid])
                    self.reporter.display(mess)

                self.dict_id[mid] = mol_name        # id --> name
                if mol_name not in self.dict_name.keys():
                    self.dict_name[mol_name] = [mid]
                # name --> list of ids
                elif mid not in self.dict_name[mol_name]:
                    self.dict_name[mol_name].append(mid)

        #print "dict_name : ",len(self.dict_name.keys())
        #print "dict_id : ",len(self.dict_id.keys())

    def make_mev_dict(self):
        """
        Since several events with same inputs and outputs (and different conditions) may exist in PID, we merge them
        """
        id_list = []
        int_list = self.page.find('Model/InteractionList')
        print len(int_list)
        mev_dict = dict()
        for inter in int_list:
            source = inter.find('Source')
            source_id = source.get('id')
            if source_id == '5': # PID curated : 5, Biocarta : 2, Reactome : 7
                self.int_cpt +=1
                int_id = inter.get('id')
                int_type = inter.get('interaction_type')
                if int_id in id_list :
                    continue # redondancy may happend when PID files are merged

                id_list.append(int_id)
                (ili, oli, acl, inl, pmidl) = self.extract_int(inter)
                mol = len(ili) + len(oli) + len(acl) + len(inl)
                self.mol_cpt += mol
                if len(ili) == 0 or len(oli) == 0: # eliminate macros
                    continue
                mev = MEvent(ili, oli, acl, inl, pmidl, self.ai_interpretation)
                try:
                    mev_p = mev_dict[mev.key]
                    mev_p.merge(mev)
                except KeyError: # new merged event
                    mev.set_clock()
                    mev_dict[mev.key] = mev
        return mev_dict

    def extract_int(self, inter):
        """
        Extract informations from a PID interaction
        """
        input_list = []
        output_list = []
        activator_list = []
        inhibitor_list = []
        pmid_list = []
        int_type = inter.get('interaction_type')

        #PMID part
        int_references = inter.find('ReferenceList')
        if int_references is not None :
            for ref in int_references :
                pmid_list.append(ref.get('pmid'))

        #Component part
        int_components = inter.find('InteractionComponentList')
        for comp in int_components :
            c_role = comp.get('role_type')
            c_name = self.throught_component(comp) # do loc and ptm mangling
            self.make_new_node(c_name, 'simple')
            if c_role == 'input' :
                input_list.append(c_name)
            elif c_role == 'output' :
                output_list.append(c_name)
            elif c_role == 'agent':
                activator_list.append(c_name)
            else :
                inhibitor_list.append(c_name)

        if int_type == 'transcription' :
            if  len(output_list)!=0 :
                gene_name = output_list[0]+'_gene'
                self.make_new_node(gene_name, 'perm')
                self.transcription_count += 1
                input_list.append(gene_name)

        return (input_list, output_list, activator_list,
                inhibitor_list, pmid_list)


    def throught_component(self, component) :
        """
        Cadbiom names result from a mangling of basic name ('PF' one if any) and various informations
        such that localisation, activity, ptm ... This is done here.
        """
        activity_state = ''
        location = ''
        ptm_name = ''

        ident = component.get('molecule_idref')
        name = self.dict_id[ident]
        label_list = component.findall('Label')
        for label in label_list :
            ltype = label.get('label_type')
            lvalue = label.get('value')
            if ltype == 'location' :
                location = self.location_converter(lvalue)
            else :
                activity_state = '_'+lvalue

        ptm_list = component.find('PTMExpression')
        if ptm_list is not None:
            for ptm in ptm_list :
                ptm_current_name = ptm_converter(ptm.get('modification'))
                ptm_name += ptm_current_name

        final_name = make_name(name, activity_state, ptm_name, location)
        final_name = make_new_name(final_name)

        return final_name

    def make_transition(self, mev):
        """
        mev is the current merge event
        """
        in_list = mev.inputs
        out_list = mev.outputs
        # condition from activators and inhibitors (ai)
        ai_condition = mev.condition

        # building guarded transitions
        for input in in_list :
            # transition is conditioned by presence of other inputs
            other_in_list = []
            for input2 in in_list :
                if input2 != input and input2 not in other_in_list :
                    other_in_list.append(input2)
            input_logic_and = logical_and(other_in_list)
            if input_logic_and and ai_condition:
                condition = '('+input_logic_and +') and ('+ ai_condition + ')'
            elif input_logic_and:
                condition = input_logic_and
            else:
                condition = ai_condition
            # generate a guarded transition from input to each output of mev
            for output in out_list :
                if input != output :
                    if not self.has_clock:
                        self.no_clock_trans(mev, input, output, condition)
                    else:
                        self.clock_trans(mev, input, output, condition)
                else:
                    #TODO warn??
                    pass

    def no_clock_trans(self, mev, input, output, condition):
        """
        translation without clock - data flow interpretation
        """
        inout = input + '#' + output
        if self.dict_transition.has_key(inout):
            # a transition already exist input -> output
            prev_transition = self.dict_transition[inout]
            previous_condition = prev_transition.condition
            # both different conditions
            cond2 = previous_condition != ''
            cond2 = cond2 and previous_condition != condition
            if condition is not None and cond2:
                new_condition = previous_condition+' or ('+condition+')'
            # equal conditions
            elif condition is not None and previous_condition != '':
                new_condition = condition
            # first condition only
            elif condition:
                new_condition = condition
            # second condition only ( may be empty chain
            else:
                new_condition = previous_condition
            prev_transition.set_condition(new_condition)
            prev_transition.note += str(mev.pmid)
        else :
            # new transition
            input_node = self.dict_node[input]
            output_node = self.dict_node[output]
            transition = self.top.add_transition(input_node, output_node)
            self.transition_count += 1
            if condition is not None :
                transition.set_condition(condition)
            transition.note = 'PMID ' + str(mev.pmid)
            self.dict_transition[inout] = transition


    def clock_trans(self, mev, input, output, condition):
        """
        translation with clock introduction
        """
        inout = input + '#' + output
        if self.dict_transition.has_key(inout):
            # a transition already exist input -> output
            prev_transition = self.dict_transition[inout]
            previous_condition = prev_transition.condition # might be none
            previous_clock = prev_transition.event

            if mev.clock == previous_clock:
                cond2 = previous_condition != ''
                cond2 = cond2 and previous_condition != condition
                if condition is not None and cond2:
                    new_condition = '(' + previous_condition + ')'
                    new_condition =  new_condition + 'or ('  + condition + ')'
                # equal conditions
                elif condition is not None and previous_condition != '':
                    new_condition = condition
                # first condition only
                elif condition:
                    new_condition = condition
                # second condition only ( may be empty chain
                else:
                    new_condition = previous_condition
                new_clock = mev.clock
            else:
                new_condition = ''
                ncl1 = when_clock(previous_clock, previous_condition)
                ncl2 = when_clock(mev.clock, condition)
                new_clock = default_clock(ncl1, ncl2)
            prev_transition.event = new_clock
            prev_transition.condition = new_condition
            prev_transition.note += str(mev.pmid)
        else:
            # new transition
            input_node = self.dict_node[input]
            output_node = self.dict_node[output]
            transition = self.top.add_transition(input_node, output_node)
            self.transition_count += 1
            if condition is not None :
                transition.set_condition(condition)
            transition.event = mev.clock
            transition.note = 'PMID '+str(mev.pmid)
            self.dict_transition[inout] = transition


    def location_converter(self, location):
        """
        convert PID location coding into cadbiom one
        """
        if self.location_dict.has_key(location) :
            self.location_dict[location] += 1
        else :
            self.location_dict[location] = 1
        # statistics
        if location in self.extern_loc:
            self.extern_count += 1
        # short hands for locations
        if location == 'transmembrane':
            return '_transMb'
        elif location =='cytoplasm' or location == 'cytosol':
            return '_cy'
        elif location == 'mitochondria':
            return  '_mi'
        elif location == 'nucleus':
            return '_nucl'
        elif location == 'extracellular region':
            return '_exCellRegion'
        elif location == 'vesicle':
            return '_v'
        elif location == 'calcium store':
            return '_calciumStore'
        elif location == 'endosome':
            return '_en'
        elif location == 'endoplasmic reticulum':
            return '_endoRetic'
        elif location == 'Golgi apparatus':
            return '_golgiAp'
        elif location == 'lysosome':
            return '_l'
        elif location == 'extracellular matrix':
            return '_exMatrix'
        elif location == 'plasma membrane':
            return '_plasmaMb'
        elif location == 'integral to membrane':
            return '_intToMb'
        elif location == 'cell-cell junction':
            return '_ccJct'
        elif location == 'hemidesmosome':
            return '_hd'
        elif location == 'caveola':
            return '_cav'
    #    elif location =='early endosome':
    #        return 'eend'
    #    elif location =='basement membrane':
    #        return 'bmb'
        else :
    #        print 'location exception : ',location
            return '_'+location




    def make_new_node(self, node_name, node_type):
        """
        add a new node in the model
        """
        if node_name in self.dict_node.keys():
                pass
        else:
            name = node_name
            if name != '':
                xloc = self.coord_x
                yloc = self.coord_y
                if node_type == 'simple':
                    node = self.top.add_simple_node(name, xloc, yloc)
                elif node_type == 'perm':
                    node = self.top.add_perm_node(name, xloc, yloc)
                elif node_type == 'macro':
                    node = self.top.add_macro_subnode(name, xloc, yloc,
                                                       0.20, 0.05)
                elif node_type == 'trap':
                    node = self.top.add_trap_node(xloc, yloc, name)
                else :
                    print 'node type error'
                self.node_count += 1
                self.dict_node[node_name] = node
                self.coord_inc()

    def coord_inc(self):
        """
        increment coordinates of nodes for a rectangular layout
        """
        if self.compt == 19:
            self.compt = 1
            self.coord_x = 0.01
            self.coord_y += 0.03
        else:
            self.compt += 1
            self.coord_x += 0.05

def logical_or(list):
    """
    @return: str - OR of the input list
    """
    if len(list) == 0 :
        return None
    elif len(list) == 1 :
        return list[0]
    else :
        logical_formula = ''
        for elemnt in list :
            logical_formula += elemnt + ' or '
        logical_formula = logical_formula[:-4]
#        print logical_formula
        return '('+logical_formula+')'

def when_clock(clo, cond):
    """
    generate a when literal expression
    """
    if not cond:
        return clo
    if len(cond) > 0:
        return '(' + clo + ') when (' + cond + ')'
    else:
        return clo

def default_clock(cl1, cl2):
    """
    generate a literal default clock
    """
    return '(' + cl1 + ') default (' + cl2 + ')'


def logical_and(lprop):
    """
    @return: str - AND of the input list
    """
    if len(lprop)==0 :
        return None
    elif len(lprop)==1 :
        return lprop[0]
    else :
        logical_formula = ''
        for elemnt in lprop :
            logical_formula += elemnt + ' and '
        logical_formula = logical_formula[:-5]
#        print logical_formula
        return '('+logical_formula+')'



def ptm_converter(ptm_term):
    """
    post transformation convertion from PID to Cadbiom
    """
    if ptm_term == 'phosphorylation':
        return '_p'
    elif ptm_term == 'acetylation':
        return '_ac'
    elif ptm_term == 'methylation':
        return '_meth'
    elif ptm_term == 'sumoylation':
        return '_sumo'
    elif ptm_term == 'ubiquitination':
        return '_ub'
    else:
        return '_' + ptm_term

def make_name(name, act, ptm, loc):
    """
    tgenerate a name
    """
    if len(act)!=0:
        name += act
    if len(ptm)!=0:
        name += ptm
    if len(loc)!=0:
        name += loc
    return name

def make_new_name(name):
    """
    remove unusable characters
    """
    new_name = ''
    for cha in name:
        if cha == ' ' :
            continue
        elif cha == '/' :
            cha = '_'
            new_name += cha
        elif cha == '-':
            cha = '_'
            new_name += cha

        elif cha == '+':
            cha = 'PLUS'
            new_name += cha

        elif cha == '.':
            cha = 'POINT'
            new_name += cha

        elif cha == '(' or cha == ')':
            cha = '__'
            new_name += cha
        else:
            new_name += cha
    return new_name

class MakeModelFromPidFile(PidXmlParser):
    """
    Interface class
    """
    def __init__(self, xml_file, reporter,
                 has_clock=True, ai_inter=0, e_loc=[]):
        PidXmlParser.__init__(self, xml_file, reporter,
                              has_clock, ai_inter, None, e_loc)




## Filename    : ana_visitors.py
## Author(s)   : Michel Le Borgne
## Created     : 04/2010
## Revision    : 03/2012
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux, Nolwenn Le Meur
##
"""
Visitors for guarded transition model analysis
"""

class TableVisitor(object):
    """
    Visitor used to collect action and place declarations.
    Events and conditions are not considered so implicit free clocks are
    not registered.
    The name of a place must be unique in a whole chart (no scope implemented).
    The same holds for an event declared in an action.
    Pseudo places used for inputs can be declared several time

    TYPES: state, input, emit
    OTHER TYPES: clock (used when free clocks are found)
    """
    def __init__(self, erdisplay):
        # place name -> (type, deepness)
        # type is state or input or emit (emitted events)
        self.tab_symb = dict()
        self.error_reporter = erdisplay     # reporter to display errors
        self.error = False                  # True if an error is found
        self.deep = 0                       # imbrication level of macro states

    def declare(self, name, stype):
        """Register an identifier with its type"""
        try:
            stype, deep = self.tab_symb[name]
            # error double declaration except for input
            self.error_reporter.display("Double declaration: "+name)
            self.error = True
        except KeyError:
            self.tab_symb[name] = (stype, self.deep)

    def visit_chart_model(self, model):
        """
        first step in the visit
        """
        model.get_root().accept(self)

    def visit_cstart_node(self, node):
        """
        nothing to declare
        """
        pass

    def visit_ctrap_node(self, node):
        """
        nothing to declare
        """
        pass

    def visit_csimple_node(self, node):
        """
        register the node as a state
        """
        self.declare(node.name, "state")

    def visit_cinput_node(self, node):
        """
        register the ident as input
        """
        try:
            ntype, deep = self.tab_symb[node.name]
            # double declaration allowed for input so we don't use declare
            # must be of the right type!
            if not (ntype == 'input'):
                self.error_reporter.display("Double declaration: " + node.name)
                self.error = True
            return
        except KeyError:
            # input places are put at level zero and are made unique
            self.tab_symb[node.name] = ('input', 0)

    def visit_cperm_node(self, node):
        """
        register the node as a state
        """
        self.declare(node.name, "state")

    def visit_cmacro_node(self, node):
        """
        register the node as a state
        deepness is updated
        """
        self.declare(node.name, "state")
        deepnest = self.deep
        self.deep = self.deep+1
        # sub states
        for sst in node.sub_nodes:
            sst.accept(self)
        self.deep = deepnest
        # transitions
        for gtr in node.transitions:
            for trans in gtr:
                trans.accept(self)

    def visit_ctop_node(self, node):
        """
        after registering the node, visit subnodes and transitions
        """
        self.declare(node.name, "state")
        # sub states
        for sst in node.sub_nodes:
            sst.accept(self)
        # transitions
        for gtr in node.transitions:
            for trans in gtr:
                trans.accept(self)

    def visit_ctransition(self, trans):
        """
        visit transition
        """
        # transition actions are the only declared events at this level
        eve = trans.action # ident syntax to be checked later!!
        if eve != "":
            try:
                ntype, deep = self.tab_symb[eve]
                # error bad type
                if ntype != "emit":
                    mess = "Double declaration - different type: " + eve
                    self.error_reporter.display(mess)
                    self.error = True
            except KeyError:
                self.tab_symb[eve] = ("emit", self.deep)


class FrontierVisitor(object):
    """
    Simple frontier computation for editor
    """
    def __init__(self):
        self.frontier = []

    def is_frontier(self, node):
        """
        Find if a node is a frontier node or a start node.
        """
        itr = node.incoming_trans
        if itr == []:
            return True
        else:
            for trans in itr:
                if trans.ori.is_start():
                    return True
        return False

    def visit_chart_model(self, model):
        """
        Entrance for the visitor
        """
        model.get_root().accept(self)

    def visit_csimple_node(self, node):
        """
        check if it is a frontier node
        """
        if self.is_frontier(node):
            self.frontier.append(node.name)

    def visit_cstart_node(self, node):
        """
        Nothing to do
        """
        pass

    def visit_ctrap_node(self, node):
        """
        Nothing to do
        """
        pass

    def visit_cinput_node(self, node):
        """
        Nothing to do
        """
        pass

    def visit_cmacro_node(self, node):
        """
        macros are not on the frontier
        """
        pass

    def visit_cperm_node(self, node):
        """
        Nothing to do
        """
        pass

    def visit_ctop_node(self, node):
        """
        visit subnodes
        """
        # sub states
        for sst in node.sub_nodes:
            sst.accept(self)

class IndirectFlowGraphBuilder(object):
    """
    build a networkx indirect graph of transitions
    """
    def __init__(self, nwx_graph):
        """
        @param nwx_graph:  networkx graph
        nwx_graph is populated by place nodes and transition edges
        """
        self.graph = nwx_graph

    def visit_chart_model(self, model):
        """
        entrance through top node
        """
        model.get_root().accept(self)

    def visit_csimple_node(self, node):
        """
        Nothing to do
        """
        pass

    def visit_cstart_node(self, node):
        """
        Nothing to do
        """
        pass

    def visit_ctrap_node(self, node):
        """
        Nothing to do
        """
        pass

    def visit_cinput_node(self, node):
        """
        Nothing to do
        """
        pass

    def visit_cmacro_node(self, node):
        """
        visit sub elements
        """
        for ltr in node.transitions:
            for trans in ltr:
                self.graph.add_edge(trans.ori.name, trans.ext.name)
        # sub states
        for sst in node.sub_nodes:
            sst.accept(self)

    def visit_cperm_node(self, node):
        """
        Nothing to do
        """
        pass

    def visit_ctop_node(self, node):
        """
        same as macro
        """
        self.visit_cmacro_node(node)

class DirectFlowGraphBuilder(object):
    """
    build a networkx graph of transitions
    """
    def __init__(self, nwx_graph, direct=True):
        """
        @param nwx_graph:  networkx Digraph
        @param direct: true if direct graph is built (default), false for the inverse graph
        nwx_graph is populated by place nodes and transition edges
        """
        self.graph = nwx_graph
        self.direction = direct

    def visit_chart_model(self, model):
        """
        entrance through top node
        """
        model.get_root().accept(self)

    def visit_csimple_node(self, node):
        """
        Nothing to do
        """
        pass

    def visit_cstart_node(self, node):
        """
        Nothing to dot
        """
        pass

    def visit_ctrap_node(self, node):
        """
        Nothing to do
        """
        pass

    def visit_cinput_node(self, node):
        """
        Nothing to do
        """
        pass

    def visit_cmacro_node(self, node):
        """
        collect edges from transitions
        """
        for ltr in node.transitions:
            for trans in ltr:
                if self.direction:
                    self.graph.add_edge(trans.ori.name, trans.ext.name)
                else:
                    self.graph.add_edge(trans.ext.name, trans.ori.name)
        # sub states
        for sst in node.sub_nodes:
            sst.accept(self)

    def visit_cperm_node(self, node):
        """
        Nothing to do
        """
        pass

    def visit_ctop_node(self, node):
        """t
        same as macro
        """
        self.visit_cmacro_node(node)


class EstimExpVisitor(object):
    """
    Partial Evaluator of event and condition expressions (subset of sig expressions)
    The value of an expression is either True (1) or False (-1) or Indeterminate (0)
    If no symbol table, all variables are evaluated to False
    """
    def __init__(self, symb_t):
        """
        @param symb_t: symbol table associating values to SOME places
        """
        self.symb_table = symb_t # this symbol table must contains name-> value

    def visit_sig_ident(self, iex):
        """
        @param iex: the biosignal ident expression
        """
        if not self.symb_table:
            return -1
        try:
            val = self.symb_table[iex.name] # 1 or -1
        except KeyError:
            val = 0 # indeterminate
        return val

    def visit_sig_const(self, cex):
        """
        coding of the constant on 1,-1
        """
        if cex.value:
            return 1
        else:
            return -1

    def visit_sig_default(self, dex):
        """
        this evaluation is valid for clocks only
        """
        cv1 = dex.left_h.accept(self)
        cv2 = dex.right_h.accept(self)
        return max(cv1 , cv2)

    def visit_sig_when(self, wex):
        """
        left operand is a clock and right operand a condition
        """
        cv1 = wex.left_h.accept(self)
        cv2 = wex.right_h.accept(self)
        return min(cv1 , cv2)

    def visit_sig_equal(self, eex):
        """
        not yet implemented
        """
        pass

    def visit_sig_sync(self, sex):
        """
        estimation depends on the operator
        """
        cv1 = sex.left_h.accept(self)
        cv2 = sex.right_h.accept(self)
        if sex.operator == 'or':
            return max(cv1 , cv2)
        if sex.operator == 'and':
            return min(cv1 , cv2)

    def visit_sig_not(self, nex):
        """
        opposite in our coding
        """
        cv1 = nex.operand.accept(self)
        return -cv1


class SigExpIdCollectVisitor(object):
    """
    Collect idents in a sig expression
    """
    def __init__(self):
        """
        simple walker
        """
        pass

    def visit_sig_ident(self, iexp):
        """
        @param iexp: the expression
        """
        return [iexp.name]

    def visit_sig_const(self, cexp):
        """
        no id
        """
        return []

    def visit_sig_default(self, dexp):
        """
        ids come from both operands
        """
        ll1 = dexp.left_h.accept(self)
        ll2 = dexp.right_h.accept(self)
        return ll1 + ll2

    def visit_sig_when(self, wexp):
        """
        ids come from both operands
        """
        ll1 = wexp.left_h.accept(self)
        ll2 = wexp.right_h.accept(self)
        return ll1 + ll2

    def visit_sig_equal(self, eex):
        """
        not implemented
        """
        pass

    def visit_sig_sync(self, sex):
        """
        ids come from both operands
        """
        ll1 = sex.left_h.accept(self)
        ll2 = sex.right_h.accept(self)
        return ll1 + ll2

    def visit_sig_not(self, nex):
        """
        same as operand
        """
        ll1 = nex.operand.accept(self)
        return ll1


